@@ -25,11 +25,15 @@
 
 #include "adw-back-button-private.h"
 #include "adw-bin.h"
+#include "adw-bottom-sheet-private.h"
+#include "adw-dialog.h"
 #include "adw-enums.h"
+#include "adw-floating-sheet-private.h"
 #include "adw-gizmo-private.h"
 #include "adw-navigation-split-view.h"
 #include "adw-navigation-view.h"
 #include "adw-overlay-split-view.h"
+#include "adw-sheet-controls-private.h"
 #include "adw-widget-utils-private.h"
 
 /**
@@ -46,6 +50,15 @@
  * features compared to it. Refer to `GtkHeaderBar` for details. It is typically
  * used as a top bar within [class@ToolbarView].
  *
+ * ## Dialog Integration
+ *
+ * When placed inside an [class@Dialog], `AdwHeaderBar` will display the dialog
+ * title intead of window title. It will also adjust the decoration layout to
+ * ensure it always has a close button and nothing else. Set
+ * [property@HeaderBar:show-start-title-buttons] and
+ * [property@HeaderBar:show-end-title-buttons] to `FALSE` to remove it if it's
+ * unwanted.
+ *
  * ## Navigation View Integration
  *
  * When placed inside an [class@NavigationPage], `AdwHeaderBar` will display the
@@ -54,9 +67,10 @@
  * When used together with [class@NavigationView] or [class@NavigationSplitView],
  * it will also display a back button that can be used to go back to the previous
  * page. The button also has a context menu, allowing to pop multiple pages at
- * once, potentially across multiple navigation views. In rare scenarios, set
- * [property@HeaderBar:show-back-button] to `FALSE` to disable the back button
- * if it's unwanted (e.g. in an extra header bar on the same page).
+ * once, potentially across multiple navigation views.
+ *
+ * Set [property@HeaderBar:show-back-button] to `FALSE` to disable this behavior
+ * in rare scenarios where it's unwanted.
  *
  * ## Split View Integration
  *
@@ -147,8 +161,8 @@ struct _AdwHeaderBar {
   GtkWidget *title_label;
   GtkWidget *title_widget;
 
-  GtkWidget *start_window_controls;
-  GtkWidget *end_window_controls;
+  GtkWidget *start_controls;
+  GtkWidget *end_controls;
   GtkWidget *back_button;
 
   char *decoration_layout;
@@ -159,11 +173,12 @@ struct _AdwHeaderBar {
   guint track_default_decoration : 1;
 
   AdwCenteringPolicy centering_policy;
-  gboolean is_mobile_window;
 
   GtkSizeGroup *size_group;
 
   GtkWidget *title_navigation_page;
+  GtkWidget *dialog;
+  GtkWidget *sheet;
 
   GSList *split_views;
 };
@@ -208,12 +223,33 @@ update_box_visibility (GtkWidget *box)
 }
 
 static void
-create_start_window_controls (AdwHeaderBar *self)
+update_decoration_layout (AdwHeaderBar *self,
+                          gboolean      start,
+                          gboolean      end)
 {
-  GtkWidget *controls = gtk_window_controls_new (GTK_PACK_START);
-  g_object_bind_property (self, "decoration-layout",
-                          controls, "decoration-layout",
-                          G_BINDING_SYNC_CREATE);
+  if (start && self->start_controls) {
+    g_object_set (self->start_controls,
+                  "decoration-layout", self->decoration_layout,
+                  NULL);
+  }
+
+  if (end && self->end_controls) {
+    g_object_set (self->end_controls,
+                  "decoration-layout", self->decoration_layout,
+                  NULL);
+  }
+}
+
+static void
+create_start_controls (AdwHeaderBar *self)
+{
+  GtkWidget *controls;
+
+  if (self->sheet)
+    controls = adw_sheet_controls_new (GTK_PACK_START);
+  else
+    controls = gtk_window_controls_new (GTK_PACK_START);
+
   g_object_bind_property (controls, "empty",
                           controls, "visible",
                           G_BINDING_SYNC_CREATE | G_BINDING_INVERT_BOOLEAN);
@@ -221,16 +257,21 @@ create_start_window_controls (AdwHeaderBar *self)
                             G_CALLBACK (update_box_visibility),
                             self->start_box);
   gtk_box_prepend (GTK_BOX (self->start_box), controls);
-  self->start_window_controls = controls;
+  self->start_controls = controls;
+
+  update_decoration_layout (self, TRUE, FALSE);
 }
 
 static void
-create_end_window_controls (AdwHeaderBar *self)
+create_end_controls (AdwHeaderBar *self)
 {
-  GtkWidget *controls = gtk_window_controls_new (GTK_PACK_END);
-  g_object_bind_property (self, "decoration-layout",
-                          controls, "decoration-layout",
-                          G_BINDING_SYNC_CREATE);
+  GtkWidget *controls;
+
+  if (self->sheet)
+    controls = adw_sheet_controls_new (GTK_PACK_END);
+  else
+    controls = gtk_window_controls_new (GTK_PACK_END);
+
   g_object_bind_property (controls, "empty",
                           controls, "visible",
                           G_BINDING_SYNC_CREATE | G_BINDING_INVERT_BOOLEAN);
@@ -238,7 +279,9 @@ create_end_window_controls (AdwHeaderBar *self)
                             G_CALLBACK (update_box_visibility),
                             self->end_box);
   gtk_box_append (GTK_BOX (self->end_box), controls);
-  self->end_window_controls = controls;
+  self->end_controls = controls;
+
+  update_decoration_layout (self, FALSE, TRUE);
 }
 
 static void
@@ -248,7 +291,7 @@ create_back_button (AdwHeaderBar *self)
 
   gtk_box_insert_child_after (GTK_BOX (self->start_box),
                               button,
-                              self->start_window_controls);
+                              self->start_controls);
   g_signal_connect_swapped (button, "notify::visible",
                             G_CALLBACK (update_box_visibility),
                             self->start_box);
@@ -285,14 +328,14 @@ update_start_title_buttons (AdwHeaderBar *self)
     }
   }
 
-  if ((self->start_window_controls != NULL) == show)
+  if ((self->start_controls != NULL) == show)
     return;
 
   if (show) {
-    create_start_window_controls (self);
-  } else if (self->start_box && self->start_window_controls) {
-    gtk_box_remove (GTK_BOX (self->start_box), self->start_window_controls);
-    self->start_window_controls = NULL;
+    create_start_controls (self);
+  } else if (self->start_box && self->start_controls) {
+    gtk_box_remove (GTK_BOX (self->start_box), self->start_controls);
+    self->start_controls = NULL;
   }
 
   update_box_visibility (self->start_box);
@@ -327,14 +370,14 @@ update_end_title_buttons (AdwHeaderBar *self)
     }
   }
 
-  if ((self->end_window_controls != NULL) == show)
+  if ((self->end_controls != NULL) == show)
     return;
 
   if (show) {
-    create_end_window_controls (self);
-  } else if (self->end_box && self->end_window_controls) {
-    gtk_box_remove (GTK_BOX (self->end_box), self->end_window_controls);
-    self->end_window_controls = NULL;
+    create_end_controls (self);
+  } else if (self->end_box && self->end_controls) {
+    gtk_box_remove (GTK_BOX (self->end_box), self->end_controls);
+    self->end_controls = NULL;
   }
 
   update_box_visibility (self->end_box);
@@ -358,6 +401,9 @@ update_title (AdwHeaderBar *self)
   if (self->title_navigation_page)
     title = adw_navigation_page_get_title (ADW_NAVIGATION_PAGE (self->title_navigation_page));
 
+  if (!title && self->dialog)
+    title = adw_dialog_get_title (ADW_DIALOG (self->dialog));
+
   if (!title) {
     GtkRoot *root = gtk_widget_get_root (GTK_WIDGET (self));
 
@@ -405,11 +451,20 @@ adw_header_bar_root (GtkWidget *widget)
   GTK_WIDGET_CLASS (adw_header_bar_parent_class)->root (widget);
 
   self->title_navigation_page =
-    adw_widget_get_ancestor_same_native (widget, ADW_TYPE_NAVIGATION_PAGE);
+    adw_widget_get_ancestor (widget, ADW_TYPE_NAVIGATION_PAGE, TRUE, TRUE);
+
+  self->dialog = adw_widget_get_ancestor (widget, ADW_TYPE_DIALOG, TRUE, FALSE);
+
+  self->sheet = adw_widget_get_ancestor (widget, ADW_TYPE_BOTTOM_SHEET, TRUE, FALSE);
+  if (!self->sheet)
+    self->sheet = adw_widget_get_ancestor (widget, ADW_TYPE_FLOATING_SHEET, TRUE, FALSE);
 
   if (self->title_navigation_page) {
     g_signal_connect_swapped (self->title_navigation_page, "notify::title",
                               G_CALLBACK (update_title), widget);
+  } else if (self->dialog) {
+    g_signal_connect_swapped (self->dialog, "notify::title",
+                              G_CALLBACK (update_title), widget);
   } else {
     GtkRoot *root = gtk_widget_get_root (widget);
 
@@ -473,6 +528,7 @@ adw_header_bar_root (GtkWidget *widget)
 
   update_title (self);
   update_title_buttons (self);
+  update_decoration_layout (self, TRUE, TRUE);
 }
 
 static void
@@ -484,13 +540,18 @@ adw_header_bar_unroot (GtkWidget *widget)
   if (self->title_navigation_page) {
     g_signal_handlers_disconnect_by_func (self->title_navigation_page,
                                           update_title, widget);
-
-    self->title_navigation_page = NULL;
+  } else if (self->dialog) {
+    g_signal_handlers_disconnect_by_func (self->dialog,
+                                          update_title, widget);
   } else {
     g_signal_handlers_disconnect_by_func (gtk_widget_get_root (widget),
                                           update_title, widget);
   }
 
+  self->title_navigation_page = NULL;
+  self->dialog = NULL;
+  self->sheet = NULL;
+
   for (l = self->split_views; l; l = l->next) {
     SplitViewData *data = l->data;
 
@@ -792,8 +853,6 @@ adw_header_bar_init (AdwHeaderBar *self)
   self->size_group = gtk_size_group_new (GTK_SIZE_GROUP_HORIZONTAL);
 
   construct_title_label (self);
-  create_start_window_controls (self);
-  create_end_window_controls (self);
   create_back_button (self);
 }
 
@@ -1180,6 +1239,8 @@ adw_header_bar_set_decoration_layout (AdwHeaderBar *self,
   if (!g_set_str (&self->decoration_layout, layout))
     return;
 
+  update_decoration_layout (self, TRUE, TRUE);
+
   g_object_notify_by_pspec (G_OBJECT (self), props[PROP_DECORATION_LAYOUT]);
 }
 
